/*	This file is part of the auxiliaries library.
	Written by Dick Grune, Vrije Universiteit, Amsterdam.
	$Id: ascii.c,v 1.2 2017-02-04 16:58:53 dick Exp $
*/

#include	"ascii.h"

/* There is little point in introducing the lib/utf8 module here */

static int
UTF8_length(const char *buf, const char *buf_end) {
	/* at least one byte present */
	int ch1 = (*buf++ & 0xFF);
	int len = 1;				/* first byte is included */
	while (ch1 & 0x40) {
		if (buf == buf_end) return -1;		/* broken UTF-8 */
		if (((*buf++ & 0xFF) & 0xC0) != 0x80) return 0;
		len++, ch1 <<= 1;
	}
	if (len < 2) return 0;
	return len;
}

const char *
conv_ascii(const char *buf, const char *buf_end, char *str) {
	char *sp = str;
	unsigned int ch= (*buf++ & 0377);

	switch (ch) {
		int i;
	case '\t':
	case '\n':
		*sp++ = ch;
		break;
	case '\b':
		*sp++ = '\\';
		*sp++ = 'b';
		break;
	case '\f':
		*sp++ = '\\';
		*sp++ = 'f';
		break;
	case '\033':
		*sp++ = '\\';
		*sp++ = 'e';
		break;
	case '\0':
		*sp++ = '\\';
		*sp++ = '0';
		break;
	case 0177:
		*sp++ = '\\';
		*sp++ = '?';
		break;
	default:
		if (' ' <= ch && ch <= '~') {
			/* ASCII96 */
			*sp++ = ch;
		}
		else if ((i = UTF8_length(buf-1, buf_end)) != 0) {
			if (i > 0) {
				*sp++ = ch;
				while (--i) {
					*sp++ = *buf++;
				}
			} else {
				/* broken UTF-8 at end of buffer */
			}
		}
		else {	/* octal */
			*sp++ = '\\';
			for (i = 2; i >= 0; i--) {
				*sp++ = ((ch >> (3*i)) & 07) + '0';
			}
		}
	}

	*sp = '\0';
	return buf;
}

