/*	This file is part of the file survey program EH.
	Written by Dick Grune, Vrije Universiteit, Amsterdam.
	$Id: eh.c,v 1.10 2017-12-14 10:11:27 dick Exp $
*/

#define	MAX_NL		4		/* Max # lines */
#define	MAX_STRING	320		/* Max # chars */
#define	MAX_YECH	16		/* Max # dirty chars */

#define	MSGUSE	"Usage is: eh [-l -p -v] file ...\n"

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<dirent.h>
#include	<errno.h>

#ifdef	MSDOS
#include	<fcntl.h>
#endif

#include	"ascii.h"
#include	"ls.h"
#include	"fname.h"

static char *errno2str(void);
static char *my_index(char *str, char ch);

/*	I / O */
static FILE *ifile;
#define	OPEN(name)	(ifile = Fopen(str2Fname(name), "rb"))
#define	IS_OPEN()	(ifile != NULL)
#define	READ(addr,size)	(fread((char *)(addr), (int)(size), 1, ifile) == 1)
#define	GETCHAR()	(getc(ifile))
#define	SEEK(pos)	(fseek(ifile, (long)(pos), 0))
#define	CLOSE()		(fclose(ifile), ifile = NULL)

/*	S e p a r a t o r s   a n d   s u f f i x e s */
#ifdef	MSDOS

#define	NAME_SEP	'\\'
#define	PATH_SEP	';'
static char *exec_suffix[] = {".exe", ".bat", "", 0};

#else	/* UNIX */

#define	NAME_SEP	'/'
#define	PATH_SEP	':'
static char *exec_suffix[] = {"", 0};

#endif	/* MDSOS/UNIX */

/*	V e r s i o n   7   S p e c i f i c s */
#ifdef	V7

#include	<a.out.h>

#define	HDR_AOUT	struct exec
#define	SYMBOL		struct nlist

#define	IS_AOUT(x)	(  (x).a_magic == A_MAGIC1 \
			|| (x).a_magic == A_MAGIC2 \
			|| (x).a_magic == A_MAGIC3 \
			|| (x).a_magic == A_MAGIC4)
#define	TXT_OFFSET(x)	(020)
#define	DAT_OFFSET(x)	(TXT_OFFSET(x) + (x).a_text)
#define	SYM_OFFSET(x)	(DAT_OFFSET(x) + (x).a_data)

#define	RELOC(hdr)	(hdr.a_flag == 0)
#define	NO_NAMES(hdr)	(hdr.a_syms == 0)

#include	"load_syms.aux"		/* include load_syms primitives */

#define	LOAD_SYMS(hdp)	load_syms(hdp)
#define	FREE_SYMS()	free_syms()

#define NAME(n)		(syms[n].n_name)
#define TYPE(n)		(syms[n].n_type)
#define VALUE(n)	(syms[n].n_value)
#define	IS_TEXT(n)	((TYPE(n) & N_TYPE) == N_TEXT)
#define IS_FNAME(n)	((TYPE(n) & N_TYPE) == N_FN)

#include	"array.aux"		/* include array routines */
#define	FIRST_SYM()	fsym_ar()
#define	NEXT_SYM()	nsym_ar()

#include	<ar.h>

#define	ARCHIVE		/* flags presence of archive format info */
typedef int ARMAGICWORD;
#define	IS_ARMAGIC(mn)	(mn == ARMAG)
#define	AR_SIZE(h)	((h.ar_size+1)&~1)

#endif	/* V7 */

/*	S y s t e m   V   S p e c i f i c s */
#ifdef	SYS_V

#include	<a.out.h>

#define	HDR_AOUT	struct bhdr

/*	We cannot use the built-in macros, since they are based on a
	global variable with a fixed name, filhdr.
*/
#define	IS_AOUT(x)	(  (x).fmagic == OMAGIC \
			|| (x).fmagic == FMAGIC \
			|| (x).fmagic == NMAGIC \
			|| (x).fmagic == IMAGIC)
#define	TXT_OFFSET(x)	((long) sizeof (HDR_AOUT))
#define	DAT_OFFSET(x)	(TXT_OFFSET(x) + (x).tsize)
#define	SYM_OFFSET(x)	(DAT_OFFSET(x) + (x).dsize)

#define	RELOC(hdr)	((hdr).rtsize != 0 || (hdr).rdsize != 0)
#define	NO_NAMES(hdr)	((hdr).ssize == 0)
#define	SIZE_SYMS(hdr)	((hdr).ssize)

/*	We cannot use the built-in type struct nlist, since it is
	incorrect.
*/
#define	SYMBOL		char

#include	"load_syms.aux"		/* include load_syms primitives */

#define	LOAD_SYMS(hdp)	load_syms(hdp)
#define	FREE_SYMS()	free_syms()

/*	This is incredible: they designed a name list format that cannot
	be accessed normally on the given machine: 50% of the alignments
	are wrong!
*/
#define	CHAR(s, i)	((s)[i] & 0377)
#define	SHORT(s, i)	((CHAR(s, i) << 8) + CHAR(s, i+1))
#define	INT(s, i)	((SHORT(s, i) << 16) + SHORT(s, i+2))

#define NAME(s)		(&(s)[6])
#define TYPE(s)		(CHAR(s, 0))
#define VALUE(s)	(INT(s, 2))
#define	IS_TEXT(s)	((TYPE(s) & N_TYPE) == N_TEXT)
#define IS_FNAME(s)	((TYPE(s) & N_TYPE) == N_FN)

static SYMBOL *iter_sym;

static SYMBOL *
FIRST_SYM(void) {
	return iter_sym = syms;
}

static SYMBOL *
NEXT_SYM(void) {
	char *s = (char *)iter_sym;

	s += (sizeof (short) + sizeof (long) + strlen(NAME(iter_sym)) + 1);
	return (s >= (char *)syms + syms_size ? NULL :
		(iter_sym = (SYMBOL *)s)
	);
}

#include	<ar.h>

#define	ARCHIVE		/* flags presence of archive format info */
typedef int ARMAGICWORD;
#define	IS_ARMAGIC(mn)	(mn == ARMAG)
#define	AR_SIZE(h)	((h.ar_size+1)&~1)

#endif	/* SYS_V */

/*	B e r k e l e y   4 . 1   S p e c i f i c s */
#ifdef	BSD4_1

#include	<a.out.h>

#define	HDR_AOUT	struct exec
#define	SYMBOL		struct nlist

#define	IS_AOUT(hdr)	!N_BADMAG(hdr)
#define	TXT_OFFSET(hdr)	N_TXTOFF(hdr)
#define	DAT_OFFSET(hdr)	N_DATOFF(hdr)
#define	SYM_OFFSET(hdr)	N_SYMOFF(hdr)

#define	RELOC(hdr)	((hdr).a_trsize != 0 || (hdr).a_drsize != 0)
#define	NO_NAMES(hdr)	((hdr).a_syms == 0)
#define	SIZE_SYMS(hdr)	((hdr).a_syms)

/*	Reading the symbol table under 4.1 BSD is a complicated affair.
	The first part can be done by straightforward load_syms, but
	the actual names have to be brought in by hand.
*/
#include	"load_syms.aux"		/* include load_syms primitives */

static char *stab;			/* separate string table */

static char *
LOAD_SYMS(const HDR_AOUT *hdp) {
	unsigned long str_len;		/* it says 4 bytes in the manual */
	char *errmsg;

	if ((errmsg = load_syms(hdp))) return errmsg;

	/* Read the actual strings */
	SEEK(N_STROFF(*hdp));
	if (!READ(&str_len, sizeof str_len)) return ", no str_len";

	stab = malloc((unsigned)str_len);
	if (stab == NULL) return ", no space for string table";

	SEEK(N_STROFF(*hdp));
	if (!READ(stab, str_len)) return ", no string table";

	return NULL;
}

static void
FREE_SYMS(void) {
	free_syms();

	/* Free the string table, if it exists */
	if (stab != NULL) {
		free((char*)stab);
	}
}

#define	NAME(s)		(&stab[(s)->n_un.n_strx])
#define TYPE(s)		((s)->n_type)
#define VALUE(s)	((s)->n_value)
#define	IS_TEXT(s)	((TYPE(s) & N_TYPE) == N_TEXT)
#define IS_FNAME(s)	(TYPE(s)==N_TEXT)	/* is this a bug ?? */

#include	"array.aux"		/* get array routines */
#define	FIRST_SYM()	fsym_ar()
#define	NEXT_SYM()	nsym_ar()

#include	<ar.h>

#define	ARCHIVE		/* flags presence of archive format info */
typedef char ARMAGICWORD[SARMAG];
#define	IS_ARMAGIC(mn)	(strncmp(mn, ARMAG, SARMAG) == 0)

static long
AR_SIZE(struct ar_hdr h) {
	char *p;
	long res = 0L;

	for (p = h.ar_size; p < h.ar_size + sizeof h.ar_size; p++) {
		if (*p != ' ') {
			res = res*10 + (*p - '0');
		}
	}
	return (res + 1) & ~1;
}
#endif	/* BSD4_1 */
/*	B e r k e l e y   4 . 2   S p e c i f i c s */
#ifdef	BSD4_2

#include	<a.out.h>

#define	HDR_AOUT	struct exec
#define	SYMBOL		struct nlist

#define	IS_AOUT(hdr)	!N_BADMAG(hdr)
#define	TXT_OFFSET(hdr)	N_TXTOFF(hdr)
#define	DAT_OFFSET(hdr)	N_DATOFF(hdr)
#define	SYM_OFFSET(hdr)	N_SYMOFF(hdr)

#define	RELOC(hdr)	((hdr).a_trsize != 0 || (hdr).a_drsize != 0)
#define	NO_NAMES(hdr)	((hdr).a_syms == 0)
#define	SIZE_SYMS(hdr)	((hdr).a_syms)

/*	Reading the symbol table under 4.1 BSD is a complicated affair.
	The first part can be done by straightforward load_syms, but
	the actual names have to be brought in by hand.
*/
#include	"load_syms.aux"		/* include load_syms primitives */

static char *stab;				/* separate string table */

static char *
LOAD_SYMS(const HDR_AOUT *hdp) {
	unsigned long str_len;		/* it says 4 bytes in the manual */
	char *errmsg;

	if ((errmsg = load_syms(hdp))) return errmsg;

	/* Read the actual strings */
	SEEK(N_STROFF(*hdp));
	if (!READ(&str_len, sizeof str_len)) return ", no str_len";

	stab = malloc((unsigned)str_len);
	if (stab == NULL) return ", no space for string table";

	SEEK(N_STROFF(*hdp));
	if (!READ(stab, str_len)) return ", no string table";

	return NULL;
}

static void
FREE_SYMS(void) {
	free_syms();

	/* Free the string table, if it exists */
	if (stab != NULL) {
		free((char*)stab);
	}
}

#define	NAME(s)		(&stab[(s)->n_un.n_strx])
#define TYPE(s)		((s)->n_type)
#define VALUE(s)	((s)->n_value)
#define	IS_TEXT(s)	((TYPE(s) & N_TYPE) == N_TEXT)
#define IS_FNAME(s)	(TYPE(s)==N_TEXT)	/* is this a bug ?? */

#include	"array.aux"		/* get array routines */
#define	FIRST_SYM()	fsym_ar()
#define	NEXT_SYM()	nsym_ar()

#include	<ar.h>

#define	ARCHIVE		/* flags presence of archive format info */
typedef char ARMAGICWORD[SARMAG];
#define	IS_ARMAGIC(mn)	(strncmp(mn, ARMAG, SARMAG) == 0)

static long
AR_SIZE(struct ar_hdr h) {
	char *p;
	long res = 0L;

	for (p = h.ar_size; p < h.ar_size + sizeof h.ar_size; p++) {
		if (*p != ' ') {
			res = res*10 + (*p - '0');
		}
	}
	return (res + 1) & ~1;
}
#endif	/* BSD4_2 */
/*	S o l a r i s   S p e c i f i c s */
#ifdef	SOLARIS

#include	<elf.h>

#define	HDR_AOUT	Elf32_Ehdr
#undef	HDR_AOUT	/* not yet implemented */

#define	IS_AOUT(hdr)	(strncmp((char *)hdr, ELFMAG, SELFMAG) == 0)

#endif	/* SOLARIS */

/*	M a i n   p r o g r a m */

static char l_flag = 0;
static char p_flag = 0;
static char v_flag = 0;

static void do_arg(char *name);
static void do_path(const char *name);
static int do_path_name(char pname[300], const char *name);
static void do_name(char *name);
static void do_file(char *name, const struct stat *stb);

int
main(int argc, char *argv[]) {
	char Version[40];
#ifdef	VERSION
	sprintf(Version, "Version %s", VERSION);
#else
	sprintf(Version, "Build %s, %s", __DATE__, __TIME__);
#endif
	while (argc > 1 && argv[1][0] == '-') {
		char *par = &argv[1][1];

		while (*par) {
			switch (*par++) {
			case 'l':		/* ls -l -like output */
				l_flag++;
				break;
			case 'p':		/* names looked up in $PATH */
				p_flag++;
				if (!getenv("PATH")) {
					fprintf(stderr, "eh: no $PATH\n");
					exit(1);
				}
				break;
			case 'v':		/* version number */
				v_flag++;
				break;
			default:
				fprintf(stderr, MSGUSE);
				exit(1);
				break;
			}
		}
		argc--, argv++;
	}

	if (v_flag) {
		printf("%s\n", Version);
		return 0;
	}

	while (argc > 1) {
		do_arg(argv[1]);
		argc--, argv++;
		fflush(stdout);
	}

	return 0;
}

static void
do_arg(char *name) {
	char *last_ch = &name[strlen(name)-1];
	if (*last_ch == '/' || *last_ch == '\\') {
		*last_ch = '\0';
	}
	if (p_flag && my_index(name, NAME_SEP) == NULL) {
		do_path(name);
	}
	else {
		do_name(name);
	}
}

static void
do_path(const char *name) {
	char *path = getenv("PATH");
	char *path_end;
	char pname[300];
	int found = 0;

	pname[0] = '\0';
#ifdef	MSDOS
	/* MSDOS searches working dir before following PATH */
	strcpy(pname, ".");
	found += do_path_name(pname, name);
#endif	/* MSDOS */

	do {
		int pn_len;

		/* isolate a path segment */
		path_end = my_index(path, PATH_SEP);
		if (path_end == NULL) {
			path_end = path + strlen(path);
		}
		pn_len = path_end - path;

		if (pn_len == 0) {
			/* empty entry, special case in UNIX */
 			strcpy(pname, ".");
			pn_len = 1;
		}
		else {
			/* extract path segment */
			strncpy(pname, path, pn_len);
		}
		pname[pn_len] = '\0';
		path = path_end+1;

		found += do_path_name(pname, name);
	} while (*path_end == PATH_SEP);

	if (!found) {
		printf("%s:\n\t%s\n\n", name, errno2str());
	}
}

static int
do_path_name(char pname[300], const char *name) {
	int pn_len = strlen(pname);
	char **esp;

	/* add file name to path name */
	pname[pn_len++] = NAME_SEP;
	strcpy(&pname[pn_len], name);
	pn_len = strlen(pname);

	for (esp = &exec_suffix[0]; *esp; esp++) {
		struct stat statb;

		/* append suffix */
		strcpy(&pname[pn_len], *esp);
		/* and see if it is there */
		if (stat(pname, &statb) == 0) {
			do_file(pname, &statb);
			return 1;
		}
	}
	return 0;
}

static void
do_name(char *name) {
	struct stat statb;

	if (Stat(str2Fname(name), &statb) == 0) {
		do_file(name, &statb);
	}
	else {
		printf("%s:\n\t%s\n\n", name, errno2str());
	}
}

							/* FILE ANALYSIS */
static void regular(const char *name);
static int text(const char *name);
static void direc(char *name);
static int archive(const char *name);
static int object(const char *name);

static void
do_file(char *name, const struct stat *stb) {
	ls(name, stb, l_flag ? "l" : "");
	printf(":\n");

	switch (stb->st_mode & S_IFMT) {
	case S_IFREG:
		regular(name);
		break;
	case S_IFCHR:
		printf("\tChar special file\n");
		break;
#ifndef	MSDOS
	case S_IFBLK:
		printf("\tBlock special file\n");
		break;
#endif	/* MSDOS */
	case S_IFDIR:
		direc(name);
		break;
#ifdef	BSD4_1
	case S_IFMPC:
		printf("\tMultiplexed char file\n");
		break;
	case S_IFMPB:
		printf("\tMultiplexed block file\n");
		break;
#endif	/* BSD4_1 */
	default:
		printf("\tUnknown file type\n");
		break;
	}
	printf("\n");
}

static void
regular(const char *name) {
	OPEN(name);
	if (!IS_OPEN()) {
		printf("\t%s\n", errno2str());
		return;
	}
	CLOSE();

	object(name) || archive(name) || text(name);
}

							/* TEXT */
static int
read_text(const char *name, char *buf) {
	int line_cnt = 0;
	char *buf_p = &buf[0];
	int ch;

	OPEN(name);

	while (	line_cnt < MAX_NL
	&&	buf_p  < &buf[MAX_STRING-1]
	&&	(ch = GETCHAR()) != EOF
	) {
		if (ch == '\n') line_cnt++;
		*buf_p++ = ch;
	}
	if (ferror(ifile)) {
		printf("%s:\n\t%s\n\n", name, errno2str());
	}
	CLOSE();

	return (buf_p - buf);
}

static void
write_text(const char *buf, int buf_len) {
	int at_BOL = 1;
	const char *buf_end = buf + buf_len;

	const char *buf_p = &buf[0];
	int yech_cnt = 0;
	while (buf_p < buf_end && yech_cnt < MAX_YECH) {
		if (*buf_p == '\n') {
			if (!at_BOL) {
				printf("\n");
				at_BOL = 1;
			}
			buf_p++;
		}
		else {
			if (at_BOL) {
				printf("\t");
				at_BOL = 0;
			}
			char str[10];
			buf_p = conv_ascii(buf_p, buf_end, str);
			printf("%s", str);
			if (*str == '\\' && strlen(str) > 2) yech_cnt++;
		}
	}
	if (!at_BOL) {
		putchar('\n');
		at_BOL = 1;
	}
}

static int
text(const char *name) {
	char buf[MAX_STRING];
	int buf_len = read_text(name, buf);
	write_text(buf, buf_len);

	return 1;
}

							/* DIRECTORY */
static void
direc(char *name) {
	int line_cnt = 0;
	Dir_t *dir;

	printf("\tDirectory:\n");
	dir = Opendir(str2Fname(name));		/* not a const parameter !!! */
	if (dir == 0) {
		printf("\t\t%s\n", errno2str());
	}
	else {
		Dirent_t *ent;

		while (	line_cnt < MAX_NL
		&&	(ent = Readdir(dir)) != 0
		) {
			const Fchar *e_Fname = ent->d_name;

			if (Fnamecmp(e_Fname, str2Fname(".")) == 0) continue;
			if (Fnamecmp(e_Fname, str2Fname("..")) == 0) continue;
			printf("\t\t%s\n", Fname2str(e_Fname));
			line_cnt++;
		}
		Closedir(dir);
	}
}

							/* ARCHIVE */
static int
archive(const char *name) {
#ifdef	ARCHIVE
	int line_cnt = 0;
	ARMAGICWORD magic;
	struct ar_hdr hdr;
	long pos;

	OPEN(name);
#ifdef	MSDOS
	setmode(fileno(ifile), O_BINARY);
#endif	/* MSDOS */
	if (	!READ(&magic, sizeof magic)
	||	!IS_ARMAGIC(magic)
	) {
		CLOSE();
		return 0;
	}

	printf("\tArchive:\n");
	pos = sizeof magic;
	while (	line_cnt < MAX_NL
	&&	(SEEK(pos), READ(&hdr, sizeof hdr))
	) {
		printf("\t\t%s\n", hdr.ar_name);
		line_cnt++;
		pos = pos + sizeof hdr + AR_SIZE(hdr);
	}
	CLOSE();
	return 1;
#else	/* !ARCHIVE */
	return name != name;
#endif	/* ARCHIVE */
}

							/* OBJECT */

/*	If it is an a.out file, we should like to know from which
	object file it derives.  Now this is a question with multiple
	answers, and we limit the answer to the name of the object file
	that contains the program start point.
	Possible program start names are given in the list below.
*/
#ifdef	HDR_AOUT
static char *start_name[] = {
	"_main",		/* C */
	"_m_a_i_n",		/* Pascal */
	0
};
#endif	/* HDR_AOUT */

static int
object(const char *name) {
#ifdef	HDR_AOUT
	HDR_AOUT hdr;

	OPEN(name);
#ifdef	MSDOS
	setmode(fileno(ifile), O_BINARY);
#endif	/* MSDOS */

	if (	!READ(&hdr, sizeof hdr)
	||	!IS_AOUT(hdr)
	) {
		CLOSE();
		return 0;
	}

	printf("\tObject");

#ifdef	RELOC
	if (RELOC(hdr)) {
		printf(", relocatable");
	}
	else
	if (NO_NAMES(hdr)) {
		printf(", no name list");
	}
	else {
		pr_fname(&hdr);
	}
#endif	/* RELOC */

	printf("\n");
	CLOSE();
	return 1;
#else	/* !HDR_AOUT */
	return name != name;
#endif	/* HDR_AOUT */
}

#ifdef	RELOC

/*	Through sheer force of will we have kept the code below
	machine/system/bias-free.
*/

static void
pr_fname(HDR_AOUT *hdp) {
	char *errmsg, *do_pr_fname();

	if (	(errmsg = LOAD_SYMS(hdp))
	||	(errmsg = do_pr_fname())
	) {
		printf(errmsg);
	}

	FREE_SYMS();
}

static char *
do_pr_fname(void) {
	SYMBOL *sym, *main_sym = NULL, *fln_sym = NULL;
	unsigned main_val;	/* value of symbol _main */

	/* Try to find the start name (_main or so) and its value */
	for (sym = FIRST_SYM(); sym; sym = NEXT_SYM()) {
		/*
			printf("type = 0x%x, val = 0x%lx, name = %s\n",
				TYPE(sym), VALUE(sym), NAME(sym)
			);
		*/
		if (	IS_TEXT(sym)
		&&	is_start_name(NAME(sym))
		) {
			main_sym = sym;
			main_val = VALUE(main_sym);
			break;
		}
	}
	if (main_sym == NULL) return ", no _main";

	/* Try to find out the name of the file it was defined in */
	for (sym = FIRST_SYM(); sym; sym = NEXT_SYM()) {
		if (IS_FNAME(sym)) {
			if (	fln_sym == NULL
			||	(	VALUE(fln_sym) < VALUE(sym)
				&&	VALUE(sym) <= main_val
				)
			) {
				fln_sym = sym;
			}
		}
	}
	if (fln_sym == NULL) {
		return ", no filename";
	}

	printf(" of %s", NAME(fln_sym));
	return NULL;
}

static int
is_start_name(const char *nm) {
	int i;

	for (i = 0; start_name[i]; i++) {
		if (strcmp(nm, start_name[i]) == 0) {
			return 1;
		}
	}
	return 0;
}
#endif	/* RELOC */

							/* ERROR HANDLING */
static char *
errno2str(void) {
	if (strerror(errno)) {
		return strerror(errno);
	}
	else {
		static char str[128];

		sprintf(str, "errno = %d\n", errno);
		return str;
	}
	/*NOTREACHED*/
}

							/* AUXILIARIES */
/*	for portability */
static char *
my_index(char *str, char ch) {
	while (*str) {
		if (*str == ch) return str;
		str++;
	}
	return 0;
}
